package se.umu.wsag.client; 

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.security.cert.X509Certificate;
import java.util.Calendar; 

import javax.xml.rpc.Stub;

import org.apache.axis.types.NCName; 
import org.globus.axis.util.Util;
import org.globus.delegation.DelegationConstants;
import org.globus.delegation.DelegationUtil;
import org.globus.gsi.GlobusCredential;
import org.globus.wsrf.client.ServiceURL;
import org.globus.wsrf.encoding.ObjectSerializer;
import org.globus.wsrf.impl.security.authentication.Constants;
import org.globus.wsrf.utils.AnyHelper; 
import org.apache.axis.message.MessageElement; 
import org.apache.axis.message.addressing.Address;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.axis.types.PositiveInteger;
import org.apache.axis.types.Duration;

import org.ggf.ws_agreement.AgreementTemplateType;
import org.ggf.ws_agreement.AgreementFactory;
import org.ggf.ws_agreement.AgreementContextType; 
import org.ggf.ws_agreement.AgreementType;
import org.ggf.ws_agreement.TermTreeType;
import org.ggf.ws_agreement.TermCompositorType;
import org.ggf.ws_agreement.ServiceDescriptionTermType;
import org.ggf.ws_agreement.GuaranteeTermType;
import org.ggf.ws_agreement.ServiceNameSet;
import org.ggf.ws_agreement.BusinessValueListType;
import org.ggf.ws_agreement.ServicePropertiesType;
import org.ggf.ws_agreement.VariableSetType;
import org.ggf.ws_agreement.VariableType;

import se.umu.wsag.impl.AgreementConstants;

import se.umu.reservation.FlexibleType;
import se.umu.reservation.DurationType;
import se.umu.reservation.EarliestStartType;
import se.umu.reservation.LatestStartType;
import se.umu.reservation.NumberOfCPUsType;
import se.umu.reservation.ReservationDescriptionType;
import se.umu.reservation.ReservationServiceLevelObjective;

/**  
 * @author tordsson
 *
 * Client for creating reservations - test purposes only
 */

public class CreateClient {
	
	private AgreementInitiator initiator; 
	private AgreementFactory factory; 
	private String eprFileName; 
	
	private String asdf; 
	
	private int earlyHours, lateHours; 
	
    static {
        Util.registerTransport(); 
    }
    
    public CreateClient(String serviceURI, String fileName) throws Exception {    	
		this.initiator = new AgreementInitiator();		
		this.factory = initiator.getFactory(serviceURI);
		
		this.asdf = serviceURI; 
		
        String clientDesc = "etc/wsag/client-security-config.xml"; 
        ((Stub)factory)._setProperty(
        		Constants.CLIENT_DESCRIPTOR_FILE, 
				clientDesc);    		
		
		this.eprFileName = fileName;
		this.lateHours = 0; 
		this.earlyHours = 0; 
    }
    
    public CreateClient(String serviceURI, String fileName, 
    		String lateTime, String earlyTime) throws Exception {
    
    	this(serviceURI, fileName); 
    	earlyHours = Integer.parseInt(lateTime); 
    	lateHours = Integer.parseInt(earlyTime); 
    }
    
    public AgreementTemplateType getTemplate() throws Exception {    				
		
		AgreementTemplateType templates[] = initiator.getTemplates(factory);
		
		System.out.println("Totally, we got " + templates.length + " templates"); 
		
		AgreementTemplateType tmpTemplate;
		AgreementTemplateType reservationTemplate = null;
		for(int i = 0; i < templates.length; i++) {
			
			tmpTemplate = templates[i]; 
			String name= tmpTemplate.getName().toString(); 

			System.out.println("Template.name: " + name); 			
			if("ReservationTemplate".equalsIgnoreCase(name)) {			
				System.out.println("We got the right template (reservation)"); 
				reservationTemplate = tmpTemplate; 							
			}			
		}
		
		if(reservationTemplate == null)
			throw new Exception("Template \"ReservationTemplate\" not found!");
		
		return reservationTemplate; 
    
    }
    
    public EndpointReferenceType delegateCredentials(String host) 
    throws Exception {
    	
    	System.out.println("Delegating credentials to service..."); 
    	
    	String URI = ServiceURL.getURL(host, 
    			DelegationConstants.FACTORY_PATH).toString(); 
    	
    	EndpointReferenceType epr = new EndpointReferenceType(); 
    	epr.setAddress(new Address(URI)); 
    	
    	X509Certificate[] certs =  DelegationUtil.getCertificateChainRP(epr,
    			null); 
    	
    	if(certs == null) {
    		System.err.println("getCertificateChainRP failed"); 
    		throw new Exception("getCertificateChainRP failed"); 
    	}
    	else {
    		GlobusCredential credential = GlobusCredential.getDefaultCredential(); 
    		
    		EndpointReferenceType delegationEpr =  DelegationUtil.delegate(
    				URI, credential, certs[0], true, null); 
    		
    		if(delegationEpr == null)
    			System.err.println("delegate failed..."); 
    		
    		else
    			System.out.println("Delegating credentials ... done"); 
    		
    		return delegationEpr; 
    	}
    }
    
    public AgreementType createOffer(AgreementTemplateType template, 
    		int cpus, 
			int dur, // (seconds) 
			Calendar earliest, 
			Calendar latest, 
			boolean flex) throws Exception {
    	
    	AgreementType offer = new AgreementType();
    	
    	NCName name = new NCName("ReservationOffer");      	
    	offer.setName(name);    	
    	
    	NCName reservationName = new NCName("Reservation"); 	    	
    	TermTreeType termsTree = new TermTreeType();
    	TermCompositorType all = new TermCompositorType();    
    	
    	// 1:  
    	// ServiceDescriptionTerms, these are added to an Object
    	// (ReservationDescriptionType) due to axis bug 15789
    	// resulting in only one ServiceDescriptionTerm in the Agreement
    	ReservationDescriptionType description = 
    		new ReservationDescriptionType();

    	// # CPUs
    	NumberOfCPUsType numberOfCPUsType = new NumberOfCPUsType();
    	String cpusString = "" + cpus; 
    	PositiveInteger numberOfCPUs = new PositiveInteger(cpusString);    
    	numberOfCPUsType.setNumberOfCPUs(numberOfCPUs);
    	
    	// Duration
    	DurationType durationType = new DurationType(); 
    	Duration duration = new Duration();  
    	duration.setSeconds(dur);     	    	
    	System.out.println("Duration (y:m:d) "  + duration.getYears() + ":"
			+ duration.getMonths() + ":" + duration.getDays());
    	System.out.println("Duration: (h:m:s) " + duration.getHours() + ":" 
    			+ duration.getMinutes() + ":" + duration.getSeconds());     	
    	durationType.setDuration(duration);
    	    
    	// Flexible
    	FlexibleType flexibleType = new FlexibleType(); 
    	flexibleType.setFlexible(true); 
    	    	
		description.setDuration(durationType); 
    	description.setNumberOfCPUs(numberOfCPUsType);
    	description.setFlexible(flexibleType); 
	
    	ServiceDescriptionTermType serviceDescriptionTerm = 
    		new ServiceDescriptionTermType();   
    	MessageElement descriptionElement[] = AnyHelper.toAnyArray(description); 
    	serviceDescriptionTerm.set_any(descriptionElement);
    	
       	all.setServiceDescriptionTerm(serviceDescriptionTerm);     	
       	
       	// 2: Set up service properties (just defines names for 
       	// usage in GuaranteeTerms), 
       	// "earlyStart" and "lateStart" are defined
       	ServicePropertiesType servicePropertiesType = 
       		new ServicePropertiesType();
       	servicePropertiesType.setName(reservationName); 
       	
       	VariableSetType variableSetType = new VariableSetType();
       	
       	VariableType earliestStartVariable = new VariableType();
       	earliestStartVariable.setName(new NCName("earlyStart")); 
       	earliestStartVariable.setMetric(
       			EarliestStartType.getTypeDesc().getXmlType());
       	
       	VariableType latestStartVariable = new VariableType(); 
       	latestStartVariable.setName(new NCName("lateStart"));  
       	latestStartVariable.setMetric(
       			LatestStartType.getTypeDesc().getXmlType());
       	
       	VariableType variables[] = new VariableType[] {
       		earliestStartVariable, latestStartVariable};        
       	variableSetType.setVariable(variables);       	       
       	servicePropertiesType.setVariableSet(variableSetType);
       	
       	all.setServiceProperties(servicePropertiesType); 
       	
       	// 3: GuaranteeTerms 
       	// (things are getting a bit more complicated here)
       	
    	// First, add serviceScope, BusinessValueList and Name, none of 
       	// these are used by the AgreementFactoryService in a reservation 
       	// scenario. 
    	GuaranteeTermType guaranteeTerm = new GuaranteeTermType();
    	
		ServiceNameSet serviceScope = new ServiceNameSet();
		NCName serviceNames[] = new NCName[]{reservationName}; 
    	serviceScope.setServiceName(serviceNames);
    	guaranteeTerm.setServiceScope(serviceScope);         	
    	BusinessValueListType businessValueList = new BusinessValueListType();
    	guaranteeTerm.setBusinessValueList(businessValueList);    	
    	guaranteeTerm.setName(reservationName); 
       	     	
    	// Next, create ServiceLevelObjectiveType and fill it with a
    	// ReservationServiceLevelObjectiveType
    	ReservationServiceLevelObjective SLOTerms = 
    		new ReservationServiceLevelObjective();    	
    	String terms[] = new String[2];     	    	
    	terms[0] = "earlyStart IS_AFTER " + (earliest.getTimeInMillis() / 1000);     
    	terms[1] = "lateStart IS_BEFORE " + (latest.getTimeInMillis() / 1000);  
    	        	
    	System.out.println(terms[0]);
    	System.out.println(terms[1]);
    	
    	SLOTerms.setTerm(terms);     
    	
    	// After registring a typemapping in server-config.wsdd 
    	// it works fine to pass a ReservationServiceLevelObjective
    	// in the guaranteeterm.  
        guaranteeTerm.setServiceLevelObjective(SLOTerms);
    	all.setGuaranteeTerm(guaranteeTerm); 
     	
    	// Add ServiceDescriptionTerms and GuaranteeTerms to agreement.  
    	termsTree.setAll(all); 
    	offer.setTerms(termsTree);  
    	
    	return offer;  
    	    
    }
    
    public void run() throws Exception {
    	
    	AgreementTemplateType reservationTemplate = getTemplate();
    	    	
    	int cpus = 1; // The old choice of 4 was out of scope for my laptop :-)
    	int duration = 3600; // 1h     	
    	Calendar earliestStart = Calendar.getInstance();
    	Calendar latestStart = Calendar.getInstance();
    	
    	if(this.earlyHours != 0) {
    		System.out.println("Using custom start time"); 
    		earliestStart.add(Calendar.HOUR, this.earlyHours);
    	}
    	else  
    		earliestStart.add(Calendar.HOUR, 2);    		
    	

    	if(this.lateHours != 0) {
    		System.out.println("Using custom start time"); 
    		latestStart.add(Calendar.HOUR, this.lateHours);
    	}
    	else
    		latestStart.add(Calendar.HOUR, 10);
    	
    	boolean flexible = false;
    	
    	System.out.println("Earliest: " + earliestStart.getTime());
    	System.out.println("Latest: " + latestStart.getTime());
    	
    	AgreementType offer = createOffer(reservationTemplate, 
    			cpus, 
				duration, 
				earliestStart, 
				latestStart, 
				flexible); 
    	// Delegerate creds and include them...
    	AgreementContextType ctx = new AgreementContextType(); 
    	EndpointReferenceType epr1 = this.delegateCredentials(this.asdf); 
    	ctx.setAgreementInitiator(epr1); 
    	offer.setContext(ctx); 
    	    	
    	EndpointReferenceType epr = initiator.createAgreement(factory, offer);
    	    	
	    String endpointString = ObjectSerializer.toString(epr,
				AgreementConstants.AGREEMENT_RESOURCE_REF);	   
	    
        FileWriter fileWriter = new FileWriter(eprFileName);
        BufferedWriter bfWriter = new BufferedWriter(fileWriter);
        bfWriter.write(endpointString);
        bfWriter.close();
        System.out.println("Endpoint reference written to file " + eprFileName);    	
           
    }

	public static void main(String[] args) {
		
		if(args.length < 1) {
			System.err.println("Usage: ReservationClient serviceURI"); 
			System.exit(0); 
		}
		
		try {
			CreateClient client = null; 
			if(args.length == 1)
				client = new CreateClient(args[0], "epr.txt");
		    if(args.length == 2)
				client = new CreateClient(args[0], args[1]);
		    if(args.length == 4)
		    	client = new CreateClient(args[0], args[1], args[2], args[3]); 
		    	client.run(); 
		} catch (Exception e) {
			System.out.println("Ooops"); 
			e.printStackTrace(); 
		}

	}
}
